use ocl::{
    builders::{BufferBuilder, KernelBuilder},
    Buffer, SpatialDims,
};
use serde::{Deserialize, Serialize};

use crate::{
    state::{Context, PROGRAM_BACKWARD, PROGRAM_FORWARD},
    Error,
};

/// A Layer of NN
pub struct Layer {
    /// vector(intermedate) for offset connection
    offset_conn: ocl::Buffer<f32>,
    /// row-base matrix(input,intermedate) for multiply connection
    mul_conn: ocl::Buffer<f32>,
    /// vector(intermedate), the output value
    ///  be processed by activation function
    output: ocl::Buffer<f32>,
    /// vector(intermedate), the output value
    /// should be processed by activation function
    pub activate: ocl::Buffer<f32>,
    /// input width
    input: usize,
    /// intermedate width
    inter: usize,
}

#[derive(Deserialize, Serialize)]
pub struct LayerDump {
    offset_conn: Vec<f32>,
    mul_conn: Vec<f32>,
}

impl Layer {
    /// dump buffers in Layer into bytes
    ///
    /// FIXME: it's blocking
    pub fn dump(self, que: &ocl::Queue) -> Result<Vec<u8>, Error> {
        let mut offset_conn = Vec::with_capacity(self.input * self.inter);
        let mut mul_conn = Vec::with_capacity(self.input * self.inter);
        // let mut param = Vec::with_capacity(self.inter);

        unsafe {
            // FIXME: should wait for queue to finish even error
            self.offset_conn
                .read(&mut offset_conn)
                .queue(que)
                .block(false)
                .enq()?;
            self.mul_conn
                .read(&mut mul_conn)
                .queue(que)
                .block(false)
                .enq()?;
            // self.param.read(&mut param).queue(que).block(false).enq()?;
        }
        que.finish()?;

        let dump = LayerDump {
            offset_conn,
            mul_conn,
            // param,
        };
        Ok(bincode::serialize(&dump)?)
    }
    /// create new Layer with random paramter
    pub fn new(input: usize, node_count: usize, state: &Context) -> Result<Self, Error> {
        let inter = node_count;

        // FIXME: should write a kernel to randomize value instead
        Ok(Self {
            offset_conn: BufferBuilder::new()
                .context(&state.context)
                .len(inter)
                .build()?,
            mul_conn: BufferBuilder::new()
                .context(&state.context)
                .len(input * inter)
                .build()?,
            output: BufferBuilder::new()
                .context(&state.context)
                .len(inter)
                .build()?,
            activate: BufferBuilder::new()
                .context(&state.context)
                .len(inter)
                .build()?,
            input,
            inter,
        })
    }
    /// forward pagination
    ///
    /// FIXME: we should use host memory instead device memory (EG. GPU)
    ///
    /// MEM_USE_HOST_PTR: use host memory, cache by device memory
    pub fn forward(&mut self, ctx: &Context, activation: &ocl::Buffer<f32>) -> Result<(), Error> {
        let kernel = KernelBuilder::new()
            .queue(ctx.queue.clone())
            .global_work_size(SpatialDims::One(self.inter))
            .program(&ctx.program[PROGRAM_FORWARD])
            .arg(&self.activate)
            .arg(&self.output)
            .arg(activation)
            .arg(&self.mul_conn)
            .arg(&self.offset_conn)
            .arg(self.input)
            .arg(self.inter)
            .build()?;

        unsafe {
            kernel.enq().unwrap();
        }

        Ok(())
    }
    /// forward pagination
    /// 
    /// delta: da superscript [L], kernel is require to rewrite it to da superscript [L-1   ]
    ///
    /// FIXME: we should use host memory instead device memory (EG. GPU)
    ///
    /// MEM_USE_HOST_PTR: use host memory, cache by device memory
    pub fn backward(&mut self, ctx: &Context, delta: &mut Buffer<f32>) -> Result<(), Error> {
        let kernel = KernelBuilder::new()
            .queue(ctx.queue.clone())
            .global_work_size(SpatialDims::One(self.inter))
            .program(&ctx.program[PROGRAM_BACKWARD])
            .arg(&self.activate)
            .arg(&self.output)
            .arg(delta)
            .arg(&self.mul_conn)
            .arg(&self.offset_conn)
            .arg(self.input)
            .arg(self.inter)
            .build()?;

        unsafe {
            kernel.enq().unwrap();
        }

        Ok(())
    }
}
